% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 8: Statistical Models

disp('Examples 8.1 and 8.2 ---------------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                  %(8.1 - 8.2)
pause;

disp('Example 8.3 ------------------------------------------------------');
clear;                                                              %(8.3a)
load datasets_ch8a
whos                                                                %(8.3b)
pause;

disp('Example 8.4 ------------------------------------------------------');
mask = trndata=='S'; % removes verse boundary delimiters	         (8.4a)
[vocab,void,index] = unique(trndata(not(mask)));
frequencies = hist(index,length(vocab)); % gets unigram counts	     (8.4b)
[unigram_count,ranking_index] = sort(frequencies,'descend');
unigram_vocab = vocab(ranking_index); % gets unigrams	             (8.4c)
unigram_probs = unigram_count/sum(unigram_count); % computes model	 (8.4d)
pause;

disp('Example 8.5 ------------------------------------------------------');
example1 = ["he","went","to","the","river","for","water"];          %(8.5a)
% computes the sentence probability according to the unigram model	%(8.5b)
prob = 1;
for k=1:length(example1)
    prob = prob * unigram_probs(unigram_vocab==example1(k));
end
disp(prob)                                                          %(8.5c)
pause;

disp('Example 8.6 ------------------------------------------------------');
sum(unigram_probs)                                                   %(8.6)
pause;

disp('Examples 8.7, 8.8 and 8.9 ----------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                 %(8.7 to 8.9)
pause;

disp('Example 8.10 -----------------------------------------------------');
freqfreq = hist(unigram_count,max(unigram_count));	                %(8.10)
pause;

disp('Example 8.11 -----------------------------------------------------');
newf(1) = 1; newff(1) = freqfreq(1); % selects singletons (c=1)	   %(8.11a)
newf(2) = 5; newff(2) = freqfreq(5); % selects the event with c=5  %(8.11b)
% selects the event with maximum number of counts c occurring twice%(8.11c)
newf(3) = max(find(freqfreq==2)); newff(3) = 2; 
% selects the event with maximum number of counts c occurring once %(8.11d)
newf(4) = max(find(freqfreq==1))+1; newff(4) = 1;
% sets slopes at the extremes for the desired smoothing function   %(8.11e)
slope1 = (log(newff(1))-log(newff(2)))/(log(newf(1))-log(newf(2)));
slopeend = 0;
% computes the smoothing function for estimating expectations 	   %(8.11f)
temp_x = log(newf); temp_y = log(newff); 
temp_xx = log(1:length(freqfreq)+1);
nc = exp(spline(temp_x,[slope1,temp_y,slopeend],temp_xx));
pause;

% Plots FIGURE 8.1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hf = figure(1); 
set(hf,'Color',[1 1 1],'Name','Frequencies of frequencies and smoothing function');
loglog(1:max(unigram_count),freqfreq,'.b',newf,newff,'dr',1:max(unigram_count)+1,nc,'.r');
legend('actual distribution','selected samples','smoothing function');
xlabel('Frequencies'); ylabel('Frequencies of frequencies');
pause

disp('Example 8.12 -----------------------------------------------------');
smoothed_counts = ((1:length(freqfreq))+1).*nc(2:end)./nc(1:end-1);%(8.12a)
unk_prob_mass = freqfreq(1)/sum(unigram_count);	                   %(8.12b)
pause;

disp('Example 8.13 -----------------------------------------------------');
unigram_scount = smoothed_counts(unigram_count);	               %(8.13a)
seen_prob_mass = (1-unk_prob_mass);                                %(8.13b)
unigram_sprobs = unigram_scount/sum(unigram_scount)*seen_prob_mass;
pause;

disp('Example 8.14 -----------------------------------------------------');
unk_prob_mass + sum(unigram_sprobs)	                                %(8.14)
pause;

disp('Example 8.15 -----------------------------------------------------');
unseenunigrams = length(unigram_vocab);	                           %(8.15a)
unk_unigram_scount = freqfreq(1)/unseenunigrams	                   %(8.15b)
unk_unigram_sprobs = unk_prob_mass/unseenunigrams                  %(8.15c)
pause;

disp('Example 8.16 -----------------------------------------------------');
unigram_sprobs(end)	                                                %(8.16)
pause;

disp('Example 8.17 -----------------------------------------------------');
absolute_diffs = unigram_probs-unigram_sprobs; 	                   %(8.17a)
relative_diffs = (unigram_probs-unigram_sprobs)./unigram_probs;    %(8.17b)
pause;

% Plots FIGURE 8.2 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hf = figure(2); 
set(hf,'Color',[1 1 1],'Name','Differences between original and discouted unigram probabilities');
loglog(1:length(absolute_diffs),absolute_diffs,'.k',1:length(relative_diffs),relative_diffs,'.b');
legend('absolute difference','relative difference','Location','northwest');
xlabel('Vocabulary rank'); ylabel('Difference between original and discounted probabilities');
pause

disp('Example 8.18 -----------------------------------------------------');
prob = 1;	                                                       %(8.18a)
for k=1:length(example1) 
    prob = prob * unigram_sprobs(unigram_vocab==example1(k));
end
disp(prob)                                                         %(8.18b)
pause;

disp('Example 8.19 -----------------------------------------------------');
find(unigram_vocab=='bob')                                          %(8.19)
pause;

disp('Example 8.20 -----------------------------------------------------');
example2 = ["bob","went","to","the","river","for","water"];	       %(8.20a)
prob = 1;                                                          %(8.20b)
for k=1:length(example2) 
    index = find(unigram_vocab==example2(k));
    if isempty(index), prob = prob * unk_unigram_sprobs;
    else, prob = prob * unigram_sprobs(index); end
end
disp(prob)                                                         %(8.20c)
pause;

disp('Examples 8.21 ----------------------------------------------------');
disp('THIS EXAMPLE IS NOT A MATLAB SCRIPT')                         %(8.21)
pause;

disp('Example 8.22 -----------------------------------------------------');
comtx = sparse(length(unigram_vocab),length(unigram_vocab));	    %(8.22)
for k=2:length(trndata) 
    w1 = trndata(k-1); w2 = trndata(k);
    if (w1~='S')&(w2~='S')
        idx1 = find(unigram_vocab==w1);
        idx2 = find(unigram_vocab==w2);
        comtx(idx1,idx2) = comtx(idx1,idx2)+1;
    end
end % coffee time!
pause; 

disp('Example 8.23 -----------------------------------------------------');
[w1,w2,rawcount] = find(comtx);                                     %(8.23)
pause;

disp('Example 8.24 -----------------------------------------------------');
[bigram_count,bigram_index] = sort(rawcount,'descend');	           %(8.24a)
bigram_word1 = unigram_vocab(w1(bigram_index));	                   %(8.24b)
bigram_word2 = unigram_vocab(w2(bigram_index));
pause;

disp('Example 8.25 -----------------------------------------------------');
% computes frequencies of frquencies                               %(8.25a)
bifreqfreq = hist(bigram_count,max(bigram_count));
% selects some points for guiding the interpolation                %(8.25b)
bif(1) = 1; biff(1) = bifreqfreq(1); 	   
bif(2) = 5; biff(2) = bifreqfreq(5);
bif(3) = max(find(bifreqfreq==2)); biff(3) = 2;
bif(4) = max(find(bifreqfreq==1))+1; biff(4) = 1;
% interpolates the smoothing function                              %(8.25c)
slope = (log(biff(1))-log(biff(2)))/(log(bif(1))-log(bif(2)));
temp_x = log(bif); temp_y = log(biff); 
temp_xx = log(1:length(bifreqfreq)+1);
binc = exp(spline(temp_x,[slope,temp_y,0],temp_xx));
% computes the smoothed count values 
temp = binc(2:end)./binc(1:end-1);
smoothed_bicounts = ((1:length(bifreqfreq))+1).*temp;              %(8.25d)
% gets the smoothed counts for seen bigrams
bigram_scount = smoothed_bicounts(bigram_count);                   %(8.25e)
pause;

disp('Example 8.26 -----------------------------------------------------');
unseenbigrams = length(unigram_vocab)^2 - length(bigram_count);	   %(8.26a)
unk_bigram_scount = bifreqfreq(1)/unseenbigrams;	               %(8.26b)
pause;

disp('Example 8.27 -----------------------------------------------------');
for k = 1:2 % computes probabilities for example1 and example2 	   %(8.27a)
    eval(sprintf('data=example%d;',k)); % selects the sentence
    prob(k) = 1; % initializes probability value to 1
    % main loop for sentence probability computation
    for n = 2:length(data) 
        % step 1: computes c(w1) 
        index1 = find(unigram_vocab==data(n-1)); % looks for w1
        % if w1 is unseen uses smoothed count for unseen unigrams
        if isempty(index1), c1 = unk_unigram_scount; 
        % if w1 is seen uses the corresponding smoothed count for w1
        else, c1 = unigram_scount(index1); end
        % step 2: computes c(w1,w2)
        word1 = bigram_word1==data(n-1); % looks for w1
        word2 = bigram_word2==data(n); % looks for w2
        index2 = find(word1&word2);
        % if w1w2 is unseen uses smoothed count for unseen bigrams
        if isempty(index2), c2 = unk_bigram_scount;
        % if w1w2 is seen uses the corresponding smoothed count for w1w2
        else, c2 = bigram_scount(index2); end
        % step 3: updates probability with p(w2|w1) = c(w1,w2)/c(w1)
        prob(k) = prob(k) * c2/c1;
    end
end
% bigram probability for example1 he went to the river for water %(8.27b)
disp(prob(1))
% bigram probability for example2 bob went to the river for water%(8.27c)
disp(prob(2))
pause;

disp('Examples 8.28 and 8.29 -------------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                %(8.28 - 8.29)
pause;

disp('Example 8.30 -----------------------------------------------------');
% step 0: initialization of variables 	                            %(8.30)
alpha = 0:0.01:1; counter = 0; logprob = zeros(1,length(alpha));
for n = 2:length(devdata) % main loop for probability computation
    % gets the following two consecutive words
    w1 = devdata(n-1); w2 = devdata(n); 
    if (w1~='S')&(w2~='S') % avoids verse boundaries
        counter = counter +1; % counts the pairs of words evaluated
        % step 1: gets unigram probabilities p(w2)
        index = find(unigram_vocab==w2); % looks for w2 
        % if w2 is unseen uses smoothed probability for unk
        if isempty(index), theprob1 = unk_unigram_sprobs;
        % if w is seen uses the corresponding smoothed unigram probability
        else, theprob1 = unigram_sprobs(index); end
        % step 2: gets bigram probabilities p(w2|w1) 
        % step 2.a: gets unigram smoothed counts c(w1) 
        index1 = find(unigram_vocab==w1); % looks for w1
        % if w1 is unseen uses smoothed count for unseen unigrams
        if isempty(index1), countw1 = unk_unigram_scount; 
        % if w1 is seen uses the corresponding smoothed count for w1
        else, countw1 = unigram_scount(index1); end
        % step 2.b: gets bigram smoothed counts c(w1,w2)
        word1 = bigram_word1==w1; % looks for w1
        word2 = bigram_word2==w2; % looks for w2
        index2 = find(word1&word2);
        % if w1w2 is unseen uses smoothed count for unseen bigrams
        if isempty(index2), countw1w2 = unk_bigram_scount;
        % if w1w2 is seen uses the corresponding smoothed count for w1w2
        else, countw1w2 = bigram_scount(index2); end
        % step 2.c: computes smoothed bigram probability
        theprob2 = countw1w2/countw1;
        % step 3: updates the log-probability sum of interpolated models
        logprob = logprob + log2((1-alpha)*theprob1 + alpha*theprob2);
    end
    % prints a dot every thousand pairs of words
    if mod(n,1000)==0, fprintf('.'); end 
end; fprintf('\n'); % this will take a while ...
pause;

disp('Example 8.31 -----------------------------------------------------');
xentropy = -logprob/counter; % computes cross-entropies	           %(8.31a)
[best_xentropy,best_alpha_index] = min(xentropy) % finds the minimum%(8.31b)
best_alpha = alpha(best_alpha_index) % gets the optimal alpha	   %(8.31c)
pause;

% Plots FIGURE 8.3 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hf = figure(3); 
set(hf,'Color',[1 1 1],'Name','Cross-entropy values in a linear combination of models');
plot(alpha,xentropy,'.-b');
xlabel('Combination parameter \alpha'); ylabel('Cross-entropy (in bits)');
text(0.02,8.75,'unigram model');
text(0.98,7.45,'bigram model','HorizontalAlign','right');
pause

% Computes TABLE 8.1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = [0.0,1.0,0.82]; 
counter = 0; logprob = zeros(1,length(alpha));
 for n = 2:length(devdata) %%%%% 8.7479    7.4593    7.0197
     w1 = devdata(n-1); w2 = devdata(n); 
%for n = 2:length(tstdata) %%%%% 8.7801    7.4971    7.0519
%    w1 = tstdata(n-1); w2 = tstdata(n); 
%for n = 2:length(trndata) %%%%% 8.7381    6.0759    6.2341
%    w1 = trndata(n-1); w2 = trndata(n);
    if (w1~='S')&(w2~='S') 
        counter = counter +1; 
        index = find(unigram_vocab==w2);  
        if isempty(index), theprob1 = unk_unigram_sprobs;
        else, theprob1 = unigram_sprobs(index); end
        index1 = find(unigram_vocab==w1); % looks for w1
        if isempty(index1), countw1 = unk_unigram_scount; 
        else, countw1 = unigram_scount(index1); end
        word1 = bigram_word1==w1; 
        word2 = bigram_word2==w2; 
        index2=find(word1&word2);
        if isempty(index2), countw1w2 = unk_bigram_scount;
        else, countw1w2 = bigram_scount(index2); end
        theprob2 = countw1w2/countw1;
        logprob = logprob + log2( (1-alpha)*theprob1 + alpha*theprob2 );
    end
    if n==1, disp('.'); elseif mod(n,1000)==0, disp(sprintf('\b.')); end; 
 end
xentropy = -logprob/counter;
pause;

disp('Examples 8.32, 8.33, 8.34 and 8.35 -------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                %(8.32 - 8.35)
pause;

disp('Example 8.36 -----------------------------------------------------');
% defines the documents                                            %(8.36a)
docs{1} = ["drink","water","and","wine"];
docs{2} = ["drink","wine","drink","water"];
docs{3} = ["eat","bread","eat","fruits"];
docs{4} = ["bread","and","fruits","are","food"];
docs{5} = ["use","spoon","and","fork"];
docs{6} = ["use","fork","use","knife"];
vocab = unique([docs{:}]) % extracts the vocabulary                %(8.36b)
pause;

disp('Example 8.37 -----------------------------------------------------');
% sets the random number generator to its default value            %(8.37a)
rng('default')
ntopics = 3; % sets the number of topics to 3                      %(8.37b)
% generates an initial guess for p(z)                              %(8.37c)
pz_ini = rand(1,ntopics); 
pz_ini = pz_ini/sum(pz_ini);
% generates an initial guess for p(w|z)                            %(8.37d)
pwz_ini = rand(length(vocab),ntopics); 
for k=1:ntopics, pwz_ini(:,k) = pwz_ini(:,k)/sum(pwz_ini(:,k)); end
pause;

disp('Example 8.38 -----------------------------------------------------');
% initializes the count matrix c(d,w)                              %(8.38a)
counts = zeros(length(docs),length(vocab));
% computes the counts of words within documents                    %(8.38b)
for k = 1:length(docs)  
    for n = 1:length(vocab)
        counts(k,n) = counts(k,n)+sum(docs{k}==vocab(n));
    end
end
pause;

disp('Example 8.39 -----------------------------------------------------');
% sets initial probabilities values                                %(8.39a)
pz = pz_ini; pwz = pwz_ini; pzd = zeros(ntopics,length(docs));
delta = 0.000001; % sets the convergence criterion threshold       %(8.39b)
% implements the simplified version of the EM algorithm            %(8.39c)
for loops=1:20
    % E-step: computes p(z|d) from p(w|z) and p(z)
    for k=1:length(docs) % loops over the documents 
        for topic=1:ntopics % loops over the topics
            temp = 1; % initializes p(d|z) 
            for n=1:length(docs{k}) % loops over words in current document
                idx = find(vocab==docs{k}(n)); % gets the word index
                temp = temp*pwz(idx,topic); % updates p(d|z)
            end
            pzd(topic,k) = temp*pz(topic); % p(z|d) <- p(d|z) p(z)
        end
        pzd(:,k) = pzd(:,k)/sum(pzd(:,k)); % normalizes p(z|d)
    end 
    % M-step: estimates p(w|z) and p(z) from p(z|d) and counts c(d,w)
    newpwz = (pzd*counts)'; % p(w|z) <- sum_d {p(z|d) c(d,w)}  
    for k=1:ntopics % normalizes p(w|z)
        newpwz(:,k) = newpwz(:,k)/sum(newpwz(:,k)); 
    end
    newpz = sum((pzd*counts)'); % p(z) <- sum_w sum_d {p(z|d) c(d,w)}
    newpz = newpz/sum(newpz); % normalizes p(z) <end of M-step>
    % computes the differences between previous and new p(w|z) and p(z)
    pwz_diff(loops) = sqrt(sum(sum((pwz-newpwz).^2)));
    pz_diff(loops) = sqrt(sum((pz-newpz).^2));
    % updates p(w|z) and p(z) with the new computed values
    pwz = newpwz; pz = newpz;
    % exits when convergence criterion is met
    if (pwz_diff(loops)<delta) && (pz_diff(loops)<delta), break; end
end
pause;

disp('Example 8.40 -----------------------------------------------------');
disp(pz_diff)                                                      %(8.40a)
disp(pwz_diff)                                                     %(8.40b)
pause;

disp('Example 8.41 -----------------------------------------------------');
% creates a table of words and words probabilities per topic        %(8.41)
wordprobs = table;
for k=1:ntopics
    [probs,idx] = sort(pwz(:,k),'descend'); 
    wordprobs.(sprintf('Topic%d',k)) = vocab(idx)';
    wordprobs.(sprintf('p(w|t%d)',k)) = compose('%4.2f',probs);
end
disp(wordprobs)
pause;

disp('Example 8.42 -----------------------------------------------------');
hf = figure(4); % creates a new figure                              %(8.42)
figure_name = 'Distribution of topics per document: p(z|d)';
set(hf,'Color',[1 1 1],'Name',figure_name);
ha = axes; bar3(pzd'); % creates the axes and the bar graph
set(ha,'XTickLabel',compose('Topic %d',1:3)); % sets axis x labels 
set(ha,'YTickLabel',compose('Doc %d',1:6)); % sets axis y labels
view([-140,35]); % rotates the view for better visualization
pause;

disp('Examples 8.43, 8.44 and 8.45 -------------------------------------');
disp('THESE EXAMPLES ARE NOT MATLAB SCRIPTS')                %(8.43 - 8.45)
pause;

disp('Example 8.46 -----------------------------------------------------');
clear; load datasets_ch8b                                          %(8.46a)
% extracts the vocabulary and selects a subset of it               %(8.46b)
[vocab,void,index] = unique(trndatatm(trndatatm~='S'));
% selects vocabulary words with frequencies in the range [2-9]
frequencies = hist(index,length(vocab));
select = (frequencies>1)&(frequencies<10);
vocab = vocab(select); 
% computes the matrix of frequency counts                          %(8.46c)
limits = find(trndatatm=='S');
counts = zeros(length(limits)-1,length(vocab));
for k = 1:length(limits)-1
    chapter = trndatatm(limits(k)+1:limits(k+1)-1);
    for n = 1:length(vocab)
        counts(k,n) = counts(k,n)+sum(chapter==vocab(n));
    end
end
pause;

disp('Example 8.47 -----------------------------------------------------');
rng('default') % resets the random number generator                %(8.47a)
ntopics = 4; % defines the number of topics                        %(8.47b)
LDAmodel = fitlda(counts,ntopics,'Verbose',0);
pause;

disp('Example 8.48 -----------------------------------------------------');
topwords = table;
for k=1:ntopics
    [probs,idx] = sort(LDAmodel.TopicWordProbabilities(:,k),'descend'); 
    topwords.(sprintf('Topic%d',k)) = vocab(idx(1:20))';
end
disp(topwords)
pause;

disp('Example 8.49 -----------------------------------------------------');
hf = figure(5); ha = axes;
figure_name = 'Distribution of topics (train documents)';
set(hf,'Color',[1 1 1],'Name',figure_name);
barh(LDAmodel.DocumentTopicProbabilities,'stacked')
xlabel("Topic probabilities")
set(ha,'YTickLabel',compose('Doc%d',1:size(counts,1)));
legend("Topic"+string(1:ntopics),...
       'Location','northoutside','Orientation','horizontal')
pause;

disp('Example 8.50 -----------------------------------------------------');
% computes the matrix of frequency counts for testdatatm            %(8.50)
limits = find(tstdatatm=='S');
counts = zeros(length(limits)-1,length(vocab));
for k = 1:length(limits)-1
    chapter = tstdatatm(limits(k)+1:limits(k+1)-1);
    for n = 1:length(vocab)
        counts(k,n) = counts(k,n)+sum(chapter==vocab(n));
    end
end
pause;

disp('Example 8.51 -----------------------------------------------------');
tsttopics = transform(LDAmodel,counts);                             %(8.51)
pause;

%%%%%%%%%% Generation of figure 8.6 
hf = figure(6); ha = axes;
figure_name = 'Distribution of topics (test documents)';
set(hf,'Color',[1 1 1],'Name',figure_name);
barh(tsttopics,'stacked')
xlabel("Topic probabilities")
set(ha,'YTickLabel',compose('Doc%d',1:size(counts,1)));
legend("Topic"+string(1:ntopics),...
       'Location','northoutside','Orientation','horizontal')
pause;


